# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# Copyright 2020 Daniel Bakkelund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

'''
Support for loading and writing ordered dissimilarity sets to and from file.

The python data structures used in the APIs in this module, are those found
in ophac.dtypes.*
'''

def _getLogger(x):
    import logging
    return logging.getLogger(__name__ + '.' + x.__name__)

class _IdMap:
    def __getitem__(self,x):
        return x

def loadDissimSpace(data=None, fname=None, keymap=_IdMap()):
    '''
    Extracts a strictly ordered dissimilarity space from the passed source.

    If data is not None:
    The dict-like must have the following contents:

    Either:
    data['N'] -> an int-like representing the number of elements in the set
    data['R'] -> array of pairs denoting order-relations in the set
    data['M'] -> array of N(N-1)/2 non-negative elements denoting the distances

    Or:
    data['Q'] -> quivers data structure (see ophac.dtypes.Quivers for details)
    data['M'] -> array of N(N-1)/2 non-negative elements denoting the distances

    If fname is not None:
    Then fname is the name of a json file that complies to the
    above description.

    Exactly one of data or fname must be specified.

    keymap - (optional) dict-like that maps either {'N','R','M'} or {'Q','M'} 
    to appropriate keys in "data".

    Returns a pair (M,Q) where Q is a ophac.dtypes.Quivers object, and M is a
    ophac.dtypes.DistMatrix object.
    '''
    import ophac.dtypes as clst
    
    assert (data is None) != (fname is None) # Exactly one must be non-None.

    if data is None:
        import json as js
        with open(fname, 'r') as inf:
            data = js.load(inf)

    Q = None
    if keymap['R'] in data:
        N = int(data[keymap['N']])
        Q = clst.Quivers(n=N,relation=data[keymap['R']])
    else:
        assert keymap['Q'] in data # Either 'Q' or 'R' must be specified
        Q = clst.Quivers(quivers=data[keymap['Q']])

    M = clst.DistMatrix(data[keymap['M']])

    assert len(Q) == M.n
    
    return (M,Q)

def dumpClustering(acs, Q=None, N=None, fname=None):
    '''
    Dumps the given clusterings to a dict.
    If a file name is specified, the result is written to file as well.
    acs   - List of AgglomerativeClustering objects
    Q     - [optional] The initial order relation
    N     - [optional] The number of elements in the space being clustered (if Q is not specified)
    fname - [optional] file to write to
    
    The json format is as follows:
    {
    "ACS":[{"joins"=[[i,j],...],
            "dists"=[d_i,...]},
            ...],
    "Q":[[..],...],
    }

    returns a dict corresponding to the json format
    '''
    import ophac.dtypes as dt
    log = _getLogger(dumpClustering)

    if Q is None:
        Q = dt.Quivers(n=N,relation=[])
    data = {
        'ACS':[{'joins':ac.joins,'dists':ac.dists} for ac in acs],
        'Q':Q.quivers
        }
    
    if fname is not None:
        import json
        log.info('Writing to file %s', fname)
        with open(fname, 'w') as outf:
            json.dump(data, outf, indent=3)

    return data



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# Copyright 2020 Daniel Bakkelund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

'''
Convert agglomerative clustering to ultrametric
'''

def _getLogger(x):
    import logging
    return logging.getLogger(__name__ + '.' + x.__name__)

def ultrametric(ac, N=-1, eps=1e-12):
    '''
    ac  - The ophac.dtypes.AgglomerativeClustering object from HC
    N   - The number of elements being clustered. If ac represents a complete
          dendrogram, N need not be specified.
    eps - The ultrametric completion threshold. Defaults to 1e-12.
    returns a ophac.dtypes.DistMatrix representing an ultrametric over the N elements.
    '''
    import ophac.dtypes as clst
    import itertools
    
    for i in range(1,len(ac)):
        if not ac.dists[i-1] <= ac.dists[i]:
            raise AssertionError('Join distances not monotone:' + \
                                     ('d[%d]=%1.4e vs d[%d]=%1.4e. (diff: %1.4e) %s' % \
                                          (i-1,ac.dists[i-1],i,ac.dists[i],ac.dists[i]-ac.dists[i-1],str(ac))))
    if N == -1:
        N = len(ac) + 1

    K = eps
    if len(ac) > 0:
        K = ac.dists[-1] + eps
        if K <= ac.dists[-1] and eps > 0:
            log = _getLogger(ultrametric)
            msg = 'Ultrametric completion threshold too low (%1.4e).' +\
                ' Lost when added to dendrogram of diameter %1.4e.'
            log.warning(msg % (eps,ac.dists[-1]))
            
    L = N*(N-1)//2
    P = clst.Partition(n=N)
    U = clst.DistMatrix([-1]*L)

    for (i,j),dist in zip(ac.joins,ac.dists):
        pi = P[i]
        pj = P[j]
        for a,b in itertools.product(pi,pj):
            if a < b:
                i1 = a
                i2 = b
            else:
                i1 = b
                i2 = a
            k = U.toLinearIndex(i1,i2)
            U.dists[k] = dist
        P = P.merge(i,j)

    basePartition = P

    while len(P) > 1:
        i,j = (len(P)-2,len(P)-1)
        for a,b in itertools.product(P[i],P[j]):
            if a < b:
                i1 = a
                i2 = b
            else:
                i1 = b
                i2 = a
            k = U.toLinearIndex(i1,i2)
            U.dists[k] = K
        P = P.merge(i,j)

    if U.max() > K: #no ultrametric...
        raise Exception('Will not produce a valid ultrametric: ' +
                        'Max ultrametric distance:%1.3f Given diam(ac) + eps = %1.3f. ' %\
                        (U.max(),K))

    if U.min() < 0: # Not all values are assigned
        raise Exception('Not all values are assigned.')

    U._basePartition = basePartition

    return U

def treeIdentical(U1,U2):
    '''
    Checks whether two ultrametrics represent the same
    hierarchical clustering in terms of partitions.
    '''
    Q1,_ = toPartitionChain(U1)
    Q2,_ = toPartitionChain(U2)
    return Q1 == Q2

def toPartitionChain(U):
    import numpy as np
    import ophac.dtypes as dt
    result = [dt.Partition(n=U.n)]
    desult = [0]
    for rho in U.spectrum(includeZero=False):
        Q = dt.Quivers([list() for _ in range(U.n)])
        for i in range(U.n):
            for j in range(i+1,U.n):
                if U[i,j] <= rho:
                    Q[i].append(j)

        Q.transitiveClosure(inPlace=True)
        done  = np.zeros((U.n,), dtype=bool)
        parts = []
        for i in range(U.n):
            if done[i]:
                continue
            else:
                part = list(Q[i])
                part.append(i)
                parts.append(sorted(part))
                for j in Q[i]:
                    done[j] = True
        
        result.append(dt.Partition(sorted(parts)))
        desult.append(rho)

    return (result,desult)

def extend(U0, P, rho, eps=1e-12):
    '''
    Extends the ultrametric by the partition P using
    the new dendrogram diameter rho.
    U0  - Ultrametric to be extended
    P   - New coarsest partition
    rho - New dendrogram diameter
    '''
    import numpy        as np
    import ophac.dtypes as dt
    
    dU0 = U0.max()
    U1  = clone(U0)
    for B in P:
        for i in range(len(B)):
            for j in range(i+1,len(B)):
                if U1[B[i],B[j]] == dU0:
                    U1[B[i],B[j]] = rho

    dists = np.array(U1.dists)
    I     = dists == dU0
    dists[I] = rho + eps

    U1 = dt.DistMatrix(list(dists))
    U1._basePartition = P
    return U1
    
    
def clone(U):
    '''
    Clones an ultrametric
    '''
    import ophac.dtypes as dt
    U2 = dt.DistMatrix(U)
    U2._basePartition = U._basePartition
    return U2



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# Copyright 2020 Daniel Bakkelund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

'''
Module for generation of random data structures.
'''

def seed(s=None):
    import time
    import numpy.random as rnd
    import random
    if s is None:
        s = int(time.time())
        rnd.seed(s)
        random.seed(s)
    else:
        rnd.seed(s)
        random.seed(s)

def randomOrderedDissimSpace(N,p,t,d1=1,steps=[1]):
    '''
    Returns a strictly ordered set of N elements where the probability
    for (i,j) in R is p, and where the expected number of ties on
    each value level is t.

    return (M,Q)
    '''
    return (randomDissimilarity(N,t,d1,steps),
            randomOrder(N,p))

def randomOrder(N,p):
    '''
    Generates an order where the initial comparability
    matrix has exactly probability p for two elements to
    be comparable; i.e. sum(Q)/(N*(N-1)/2)=p.
    '''
    import ophac.dtypes as dt
    import numpy        as np
    from numpy.random import permutation as perm
    n      = int(round(N*(N-1)//2*p))
    q      = np.random.random(size=N*(N-1)//2)
    quivs  = [list() for _ in range(N)]
    lindex = 0
    for i in range(N):
        for j in range(i+1,N):
            if q[lindex] <= p:
                quivs[i].append(j)

            lindex += 1


    pi    = np.random.permutation(np.arange(N))
    ip,_  = zip(*sorted(zip(np.arange(N),pi), key=lambda x : x[1]))
    quivs = [[pi[x] for x in quivs[ip[i]]] for i in np.arange(N)] 
            
    return dt.Quivers(quivs)

def randomDissimilarity(N,t,d1=1,steps=[1],scale=0.0):
    '''
    N     - Number of elements
    t     - The expected value of number of equal values on each
            dendrogram level following a uniform distribution.
    d1    - The lowest value in the dissimilarity matrix.
            Default: 1
    steps - Array of increments to be cycled
            Default: [1]
    scale - Standard deviaton of normal noise to add to the distances.
            Default: 0.0
    '''
    import numpy        as np
    import ophac.dtypes as dt
    import itertools

    t       = int(np.round(t))
    M       = N*(N-1)//2
    val     = float(d1)
    steps   = itertools.cycle(steps)
    result  = np.zeros((M,), dtype=float)
    k       = 0
    while k < M:
        mult = np.min([M-k,t])
        assert mult > 0
        result[k:k+mult] = val
        k   += mult
        val += next(steps)

    assert k == M
    
    result = np.random.permutation(result)

    if scale > 0:
        result += np.random.normal(0,scale,size=result.shape)
        if np.any(result < 0):
            raise Exception('Too much noise --> negative dissimilarities.')

    dists = dt.DistMatrix(list(result))
    assert N == dists.n
    return dists
    
    

    
def randomDissimilarity_old(N,n,d1=1,steps=[1],scale=0.0):
    '''
    N     - Number of elements
    n     - The expected value of number of equal values on each
            dendrogram level following a uniform distribution.
    d1    - The lowest value in the dissimilarity matrix.
            Default: 1
    steps - Array of increments to be cycled
            Default: [1]
    scale - Standard deviaton of normal noise to add to the distances.
            Default: 0.0
    '''
    from ophac.dtypes     import DistMatrix
    from numpy.random     import random      as rnd
    from numpy.random     import permutation as perm
    from numpy.random     import normal      as noise
    import itertools

    M = N*(N-1)//2
    indices = list(range(0,M))
    result  = [-1]*M
    
    incs = itertools.cycle(steps)
    val  = d1
    while len(indices) > 0:
        m = max(1,int(rnd()*n*2))
        k = min(m,M)
        L = min(len(indices),k)
        I = perm(len(indices))[:L]
        I = sorted(I, reverse=True)
        for i in I:
            result[indices[i]] = val + noise(scale=scale)
            del indices[i]
        val += next(incs)

    return DistMatrix(result)

def plantedPartition(Q,N,p,dEq=0.1,dDiff=1.0,s=0.01):
    '''
    Produces planted partitions by duplicating the ordered
    set Q times N. Two equivalent elements are then registered
    as equivalent with probability p by specifying their
    dissimilarity by dEq. Otherwise, and for all non-equivalent
    elements, they are registered with dissimilarity dDiff.
    Finally, Gaussian noise is added to the dissimilarity measure
    with mean 0 and variance s.

    Q     - Base ordered set.
    N     - Number of parallel datasets. 
            Must be >= 1.
    p     - Probability for equivalent elements to be marked so. 
            Must be 0 <= p <= 1.
    dEq   - Dissimilarity for elements marked as equivalent. 
            Must be >0.
    dDiff - Dissinilarity for all other elements. 
            Must be >dEq.
    s     - Variance of Gaussian noise to add. 
            Must be >=0.

    Returns M,Q2,P where 
    M  is the produced dissimilarity measure,
    Q' is the corresponding partial order, and
    P  is the planted partition.
    '''
    import numpy as np
    from numpy.random import random as rnd
    from numpy.random import normal 
    from itertools import combinations
    import ophac.dtypes             as dt
    
    M       = N*len(Q) # num elts in produced poset
    quivers = []
    planted = [list() for _ in range(len(Q))]
    dists   = dt.DistMatrix([dDiff]*(M*(M-1)//2))

    # Concatenate all quivers into one
    # and create the planted partition
    for n in range(N):
        dn = n*len(Q)
        
        assert len(quivers) == dn
        for quiver in Q.quivers:
            quivers.append([k+dn for k in quiver])

        for k in range(len(Q)):
            planted[k].append(k+dn)
            
    Q2 = dt.Quivers(quivers)
    assert len(Q2) == M

    # Assign dissimilarities
    for plant in planted:
        for i,j in combinations(plant, 2):
            if rnd() <= p:
                dists[i,j] = dEq

    # Add some noise
    if s > 0:
        noise = np.abs(normal(loc=0, scale=s, size=len(dists.dists)))
        dists = dt.DistMatrix(noise + dists.dists)
                
    return dists,Q2,dt.Partition(planted)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# Copyright 2020 Daniel Bakkelund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

def _getLogger(thing):
    import logging
    return logging.getLogger(__name__ + '.' + thing.__name__)

def linkage(D, G=None, L='complete', p=1, K=1.0e-12):
    '''
    Performs order preserving hierarchical clustering.
    
    D - Condensed dissimilarity measure. That is, for a set of N elements,
        D is a 1-dimensional array of N(N-1)/2 elements, corresponding to the
        upper half of the dissimilarity matrix laid out row-wise.
    G - Order relation given as a jagged 2D-array as follows:
        if A = G[i] is the i-th array in G, then, for each k in A this is 
        interpreted as i<k in the strict partial order. Default is None,
        meaning that the hierarchical clustering will be performed without
        regard to any order relation.
    L - Linkage function. One of 'single', 'complete' or 'average'.
        Default is 'complete'.
    p - Order of norm. A number greater or equal to 1. Defaults to 1.
    K - The epsilon of the ultrametric completion. Defaults to 1.0e-12.

    The method returns a list of ophac.dtypes.AgglomerativeClusering (AC) objects, 
    each object has two members: "dists" and "joins". The joins hold the order 
    of merges of the clustering, the the dists are the distances for the 
    different joins.

    The AC objects can be passed on as follows to obtain relevant information.
    In the following, ac is an AC object, N is the number of elements in the base set
    and K is the ultrametric completion distance (epsilon). Unless you have used
    a non-default value of K, this may be omitted.

    * To create the corresponding ultrametric:
    import ophaq.ultrametric as ult
    U = ult.ultrametric(ac, N, K)

    * To obtain the k-th partition
    import ophaq.dtypes as dt
    kthPartition = dt.merge(dt.Partition(n=N), ac.joins[:k])

    * To obtain the k-th induced order relation
    import ophac.dtypes as dt
    kthInducedOrder = dt.merge(dt.Quivers(G), ac.joins[:k])

    * To plot the corresponding (partial) dendrogram
    import matplotlib.pyplot as plt
    import ophaq.dendrogram  as dend
    fix,ax = plt.subplots(1,1)
    dend.plot(ac, N, ax=ax)
    plt.show()

    The  dend.plot(...) method uses scipy.cluster.hierarchy.dendrogram for 
    the plotting, and passes any additional arguments on to that method.
    '''
    import ophac.hac as hac
    import numpy     as np
    import time
    
    log = _getLogger(linkage)

    if np.min(D) < 0:
        msg = 'Negative dissimilarity coefficients detected.'
        log.error(msg)
        raise ValueError(msg)
    
    M = hac.DistMatrix(D)
    log.info('Clustering %d elements.', M.n)
    Q = None
    if G:
        Q = hac.Quivers(G)
    else:
        log.info('No order relation specified.')
        Q = hac.Quivers(relation=[], n=M.n)

    start    = time.time()
    hc       = hac.HAC(lnk=L, ord=p, dK=K)
    acs      = hc.generate(M,Q)
    duration = time.time() - start
    log.info('Produced %d equivalent clusterings in %1.3f s.', len(acs), duration)
    return acs

    
def approx_linkage(D, G=None, L='single', n=10, mode='rndpick', procs=4, p=1, K=1e-12):
    '''
    Produces an order preserving hierarchical clustering of (M,Q) using parallel 
    processing with a number of processes and a number of samples to generate.

    The algorithm applies a level of noise to the dissimilarity measure to ensure 
    there are no initial tied connections, causing each run to be a polynomial 
    time algorithm for all linkage models. The clusterings are run in parallel, 
    and the returned resulst is the set of optimal clusterings.

    For each optimal (noise perturbated) solution, a de-noised version is produced
    by re-playing the joins on the original dissimilarity measure.
    
    n      - The number of samples to run.
    procs  - Number of processes to run in parallel. Defaults to 4.
    mode   - The approximation mode to use.
             untied  - add noise to the dissimilarity and run un-tied
             rndpick - resolve ties by picking a random pair
             Default is rndpick.

    For the other parameters, se "linkage".

    Returns a list of AC objects, one for each equivalently optimal solution.
    In general, it is expeted that there is only one optimal solution, due to
    the noise added to the dissimilarities, but then again...
    '''
    import multiprocessing as mp
    import numpy           as np
    import ophac.hac       as hac
    import time

    log = _getLogger(approx_linkage)

    if np.min(D) < 0:
        msg = 'Negative dissimilarity coefficients detected.'
        log.error(msg)
        raise ValueError(msg)

    
    # Convert types to serialisable ones
    mm = D
    if isinstance(mm,hac.DistMatrix):
        mm = mm.dists

    log.info('Clustering %d elements %d times using %d processors.',
            hac.DistMatrix(mm).n, n, procs)

    qq = G
    if isinstance(qq,hac.Quivers):
        qq = G.quivers

    if qq is None:
        qq = [list()]*hac.DistMatrix(mm).n
        
    approx_method = None
    if mode == 'untied':
        approx_method = _untied_linkage
    elif mode == 'rndpick':
        approx_method = _rndpick_linkage
    else:
        raise Exception('Unknown approximation mode: "' +  mode + '"')

    CPP_MAX_UINT = 4294967294
    
    seed = np.random.randint(CPP_MAX_UINT, size=n)
    data = [(mm,qq,L,int(s)) for s in seed]
    with mp.Pool(processes=procs) as pool:
        results = pool.map(approx_method, data)

    acs      = [hac.AC(j,d) for j,d,_ in results]
    d0       = hac.DistMatrix(mm)

    result = hac._pickBest(d0, acs=acs, ord=p, dK=K)
    if len(result) > 1:
        log.warning('Random clustering resulted in %d equivalent results.', len(result))
        
    return result

def _rndpick_linkage(XX):
    import ophac.dtypes     as dt
    import ophac.hac_approx as hac
    import time
    
    mm,qq,lnk,seed = XX
    log = _getLogger(_rndpick_linkage)
    stt = time.time()
    D   = dt.DistMatrix(mm)
    Q   = dt.Quivers(qq)
    hc  = hac.HACUntied(lnk)
    ac  = hc.generate(D,Q,mode='approx',seed=seed)
    log.info('Time: %1.4f s.', time.time() - stt)
    return (ac.joins,ac.dists,hc.seed_used)

def _p_linkage(XX):
    _getLogger(_p_linkage).warning('Deprecated. use ophac.hierarchy._untied_linkage')
    return _untied_linkage(XX)

def _untied_linkage(XX):
    '''
    Parallel method called by pool.map.
    TODO: consider [gaussian] noise about zero. The current solution
          moves the mass center of the dissimilarity.
    '''
    import numpy            as np
    import ophac.hac_untied as hac
    import time
    
    mm,qq,lnk,_ = XX
    log = _getLogger(_untied_linkage)
    stt = time.time()
    
    N = hac.DistMatrix(mm).n
    
    dd0  = np.array(sorted(set(mm)), dtype=float)
    seps = dd0[1:] - dd0[:-1]
    sep  = np.min(seps)

    srt = time.time()
    rnd = np.random.random(np.shape(mm)) * (sep/(3*N))
    rnt = time.time() - srt

    Q = None
    if qq is not None:
        Q = hac.Quivers(qq)

    M   = hac.DistMatrix(list(mm + rnd))
    hc  = hac.HACUntied(lnk)
    prt = time.time() - stt
    sht = time.time()
    ac  = hc.generate(M,Q,mode='untied')
    ttm = time.time() - sht
    log.info('Time: %1.4f s. (random generation: %1.4f s, prep. time: %1.4f s.)' % \
             (ttm,rnt,prt))

    denoised = list(_dists(ac.joins,mm,lnk))
    return (ac.joins,denoised,ac.seed_used)

def _dists(joins,D,L,precision=30):
    '''
    Produces the join distances for the given joins and the dissimilarity
    for the linkage method specified.

    precision - number of decimals to round off

    Is used to produce un-noised join distancs returned by _p_linkage.
    '''
    import ophac.dtypes as dt
    import numpy        as np

    if not isinstance(D,dt.DistMatrix):
        D = dt.DistMatrix(D)
    
    dists = []
    lfact = dt.getLinkageFactory(L)
    sizes = [1]*D.n
    D1    = dt.DistMatrix(D)    
    for i,j in joins:
        lnk = lfact(sizes)
        d   = D1[i,j]
        D1  = lnk(i,j,D1)
        sizes[i] += sizes[j]
        del sizes[j]

        dists.append(d)

    return np.round(dists, decimals=precision)




# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# Copyright 2020 Daniel Bakkelund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

'''
Data structures for AHC of posets.
'''

def _getLogger(thing):
    import logging
    if isinstance(thing,str):
        return logging.getLogger(__name__ + '.' + thing)
    else:
        return logging.getLogger(__name__ + '.' + thing.__name__)

def merge(thing, joins, *args, **kwargs):
    if not isinstance(thing, DistMatrix):
        t = thing
        for i,j in joins:
            t = t.merge(i,j, *args, **kwargs)
        return t
    else:
        lnk = getLinkage(kwargs['L'])
        t = thing
        for i,j in joins:
            t = lnk(i,j,t)
        return t

class Partition:

    log = _getLogger("Partition")

    def __init__(self, data=None, n=None):
        assert (data is None) != (n is None)
        if data:
            assert len(data) > -1 # Sanity check...
            self.data = data
        else:
            self.data = [[i] for i in range(n)]

        self.log.debug('Created: %s', self.data)

    def merge(self,a,b):
        assert a < b
        data        = [None]*(len(self.data) - 1)
        data[:a]    = self.data[:a]
        data[a]     = sorted(self.data[a] + self.data[b])
        data[a+1:b] = self.data[a+1:b]
        data[b:]    = self.data[b+1:]
        return Partition(data=data)

    def __contains__(self,x):
        return x in self.data

    def __getitem__(self,i):
        return self.data[i]

    def __len__(self):
        return len(self.data)

    def __eq__(self,other):
        if not isinstance(other, Partition):
            return False
        return self.data == other.data

    def __ne__(self,other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(tuple([tuple(x) for x in self.data]))

    def __str__(self):
        return 'Partition(%s)' % str(self.data)

    def __repr__(self):
        return str(self)
    

class Quivers:
    def __init__(self, quivers=None, n=None, relation=None):
        assert ((quivers is None) != (n is None)) \
            and ((n is None) == (relation is None))
        
        if quivers:
            if isinstance(quivers, Quivers):
                self.quivers = [list(q) for q in quivers.quivers]
            else:
                assert len(quivers) > -1 # Sanity again...
                self.quivers = [sorted(set(q)) for q in quivers]
        else:
            import bisect
            self.quivers = [[] for _ in range(n)]
            for a,b in relation:
                i = bisect.bisect_left(self.quivers[a], b)
                self.quivers[a].insert(i,b)

        self.transitivelyClosed = False

    def merge(self,a,b):
        assert a < b
        quivers = [list(x) for x in self.quivers]
        quivers[a].extend(quivers[b])
        quivers[a] = sorted(set(quivers[a]))
        del quivers[b]

        # Reduce pointers beyond the deletion pont,
        # and redirect pointers to the deleted vertex by pointers
        # to the merged vertex
        for quiver in quivers:
            for i in range(len(quiver)):
                if quiver[i] == b:
                    quiver[i] = a
                elif quiver[i] > b:
                    quiver[i] -= 1
               
        result = Quivers(quivers=quivers)
        return result

    def canMerge(self,a,b):
        return not (self.hasPath(a,b) or self.hasPath(b,a))

    def hasPath(self,a,b,visited=None):
        if self.transitivelyClosed:
            import bisect
            i = bisect.bisect_left(self.quivers[a],b)
            return i < len(self.quivers[a]) and self.quivers[a][i] == b
        else:
            if visited is None:
                visited = set()

            visited.add(a)

            if a == b:
                return True
            for c in self.quivers[a]:
                if not c in visited:
                    if self.hasPath(c,b,visited):
                        return True

            return False

    def hasCycle(self, n=None, visiting=None, visited=None):
        if self.transitivelyClosed:
            import bisect
            for n in range(len(self)):
                i = bisect.bisect_left(self.quivers[n],n)
                if i < len(self.quivers[n]) and self.quivers[n][i] == n:
                    return True
            return False
        else:
            assert (n is None) == (visiting is None) == (visited is None)
            if n is None:
                visited  = [False]*len(self)
                visiting = [False]*len(self)
                for n in range(len(self.quivers)):
                    if not visited[n]:
                        if self.hasCycle(n,visiting,visited):
                            return True
                return False
            else:
                if visiting[n]:
                    return True

                visiting[n] = True
                for ch in self.quivers[n]:
                    if not visited[ch] and self.hasCycle(ch,visiting,visited):
                        return True

                visited[n]  = True
                visiting[n] = False
                return False

    def transitiveClosure(self,inPlace=False,lenient=False):
        '''
        Produces a Quivers that is transitively closed.
        If inPlace is True, this object is changed. Otherwise,
        a new Quivers object is returned.
        If lenient is True, cycles will not lead to exceptions.
        '''

        def _fillInDescendants(n, visited, visiting, visitseq, data):
            if visiting[n]:
                if not lenient:
                    raise Exception('Cycle in Quivers object detected %d -> %d' % (n,n))
                else:
                    i = visitseq.index(n)
                    decs = set.union(*data[i:])
                    decs.add(n)
                    return decs

            if visited[n]:
                return data[n]

            visiting[n] = True
            visitseq.append(n)
            decs        = data[n]
            for d in self.quivers[n]:
                decs.add(d)
                if not visited[d]:
                    decs.update(_fillInDescendants(d,visited,visiting,visitseq,data))
                else:
                    decs.update(data[d])
                
            visited[n]  = True
            visiting[n] = False
            visitseq.pop()
            #data[n]     = set(decs)
            return decs

        
        N        = len(self)
        data     = [set() for _ in range(N)]
        visited  = [False]*N
        visiting = [False]*N
        visitseq = []
        for i in range(N):
            _fillInDescendants(i,visited,visiting,visitseq,data)

        
            
        if inPlace:
            self.quivers = [sorted(d) for d in data]
            self.transitivelyClosed = True
            return self
        else:
            result = Quivers(quivers=data)
            result.transitivelyClosed = True
            return result

    def connectedComponents(self):
        '''
        Returns a list of node sets, one for each connected
        component.
        '''
        visited = set()
        result  = []
        revMap  = self._genRevMap()
        for n in range(len(self)):
            if n in visited:
                continue
            # No CC containing n:
            cc = set()
            self._addToCC(cc,n,self.quivers,revMap)
            visited.update(cc)
            result.append(cc)

        return result

    def _addToCC(self,cc,n,fwd,rev):
        cc.add(n)
        for m in rev[n].union(fwd[n]):
            if not m in cc:
                self._addToCC(cc,m,fwd,rev)
        
    def _genRevMap(self):
        import collections
        revMap = collections.defaultdict(set)
        for p in range(len(self)):
            subs = self[p]
            for s in subs:
                revMap[s].add(p)
        return [revMap.get(n,set()) for n in range(len(self))]


    def degrees(self,transitive=False):
        '''
        Returns a sequence of pairs, one pair for every
        element, containing the elements' in- and out degree.
        '''
        Q = self
        if transitive and not self.transitivelyClosed:
            Q = self.transitiveClosure(inPlace=False)

        outs = [len(x) for x in Q.quivers]
        ins  = [0]*len(Q)
        for q in Q.quivers:
            for to in q:
                ins[to] += 1
                
        return zip(ins,outs)

    def toAdjacencyMatrix(self):
        import numpy as np
        n = len(self)
        A = np.zeros((n,n), dtype=int)
        for i in range(n):
            for j in self.quivers[i]:
                A[i,j] = 1

        return A
    
    def __getitem__(self,i):
        return self.quivers[i]

    def __eq__(self,other):
        if not isinstance(other, Quivers):
            return False
        return self.quivers == other.quivers
    
    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(tuple([tuple(x) for x in self.quivers]))

    def __len__(self):
        return len(self.quivers)

    def __str__(self):
        return 'Quivers(%s)' % str(self.quivers)

    def __repr__(self):
        return str(self)

class DistMatrix:

    log = _getLogger('DistMatrix')
    
    def __init__(self, dists):
        from math import sqrt,floor
        if isinstance(dists, DistMatrix):
            self.dists   = list(dists.dists)
            self.n       = dists.n
        else:
            self.dists = dists
            n = 0.5 + 0.5 * sqrt(1 + 8*len(dists))
            if n != floor(n):
                raise AssertionError('n=%f for len(dists)=%d.' % (n,len(dists)))
            self.n = int(n)

        self.shape = [self.n,self.n]

        self.log.debug('Created %dx%d matrix.', self.n, self.n)

    @staticmethod
    def fromDissimilarity(N, dissimFunc):
        dists = []
        for i in range(N):
            for j in range(i+1,N):
                dists.append(dissimFunc(i,j))
        return DistMatrix(dists)

    def toNumpyArray(self):
        import numpy as np
        A = np.zeros((self.n,self.n), dtype=float)
        for i in range(0,self.n):
            for j in range(i+1,self.n):
                k = self.toLinearIndex(i,j)
                x = self.dists[k]
                A[i,j] = x
                A[j,i] = x

        return A

    def _getLinkage(self,name):
        if name == 'single':
            result = lambda a,b : min(a,b)
            result.__name__ = name
            return result
        if name == 'complete':
            result = lambda a,b : max(a,b)
            result.__name__ = name
            return result

        raise ValueError(str(name) + ' is not a valid linkage strategy.')

    def __getitem__(self,*args):
        if type(args[0]) is tuple:
            if args[0][0] < args[0][1]:
                a = args[0][0]
                b = args[0][1]
            else:
                a = args[0][1]
                b = args[0][0]

            if a == b:
                return 0.0

            return self.dists[self.toLinearIndex(a,b)]
        else:
            return self.dists[args[0]]

    def __setitem__(self,*args):
        if type(args[0]) is tuple:
            if args[0][0] < args[0][1]:
                a = args[0][0]
                b = args[0][1]
            else:
                a = args[0][1]
                b = args[0][0]
            x = args[1]

            if a == b:
                assert x == 0.0
                return self

            i = self.toLinearIndex(a,b)
            self.dists[i] = x
            return self
        else:
            self.dists[args[0]] = args[1]
            return self

    def spectrum(self,includeZero=False):
        result = sorted(set(self.dists))
        if includeZero:
            result.insert(0,0)
        return result

    def max(self):
        return max(self.dists)

    def min(self):
        return min(self.dists)

    def norm(self,ord=2):
        '''
        Returns the L_{ord,ord} norm of this dissimilarity as if it
        were an NxN matrix.
        '''
        import numpy.linalg as npl
        scale = 2.0**(1/float(ord))
        return scale * npl.norm(self.dists,ord=ord)

    def __add__(self,other):
        assert isinstance(other, DistMatrix)
        assert other.n == self.n
        data = [other.dists[i] + self.dists[i] for i in range(len(self.dists))]
        return DistMatrix(data)

    def __sub__(self,other):
        assert isinstance(other, DistMatrix)
        assert other.n == self.n
        data = [self.dists[i] - other.dists[i] for i in range(len(self.dists))]
        return DistMatrix(data)

    def __mul__(self,other):
        if isinstance(other,DistMatrix):
            raise Exception('Matrix multiplication not supported.')
        else:
            data = list(self.dists)
            for i in range(len(data)):
                data[i] = data[i]*other

            return DistMatrix(data)

    def __rmul__(self,other):
        return self*other

    def __div__(self,other):
        return self * (1.0/other)

    def __truediv__(self,other):
        return self * (1.0/other)

    def getSortedIndexPairs(self):
        '''
        Returns a list of pairs on the form
        (d,(i,j)), where d is the dissimilarity d(i,j).
        The entries are sorted w.r.t. accending values of d.
        '''
        import itertools
        pairs   = itertools.combinations(range(self.n),2)
        entries = zip(self.dists, pairs)
        key     = lambda a : a[0]
        return sorted(entries, key=key)

    def getChunkedIndexPairs(self):
        '''
        Returns a list of Chunk objects, each having two attributes:
        dist:  the dissimilarity of the chunked pairs
        pairs: a list of index pair tuples (i,j)
        The list is sorted in accending order, so popping elements will
        provide the next chunk of minimal links.
        '''
        pairs = self.getSortedIndexPairs()
        if len(pairs) == 0:
            return []

        class Chunk:
            def __str__(self):
                return 'Chunk[%1.3f,%s]' % (self.dist, str(self.pairs))
            def __repr__(self):
                return str(self)
            def __len__(self):
                return len(self.pairs)

        result = []
        chunk  = Chunk()
        chunk.dist  = pairs[-1][0]
        chunk.pairs = [pairs[-1][1]]
        result.append(chunk)
        for pair in pairs[-2::-1]:
            if pair[0] == chunk.dist:
                chunk.pairs.append(pair[1])
            else:
                chunk = Chunk()
                chunk.dist  = pair[0]
                chunk.pairs = [pair[1]]
                result.append(chunk)
        
        return result

    def toLinearIndex(self,*args):
        assert len(args) == 2
        i = args[0]
        j = args[1]
        assert i < j
        
        N = self.n*(self.n-1)//2        # Half matrix size
        r = (self.n-i)*(self.n-i-1)//2  # Half matrix size including and below point
        h = j-i-1                       # Number of cells to the right of the diagonal
        return N - r + h


    def toMatrixIndex(self,k):
        from math import sqrt,floor
        n = self.n
        i = n - 2 - int(floor(sqrt(-8*k + 4*n*(n-1)-7)/2.0 - 0.5))
        j = k + i + 1 - n*(n-1)//2 + (n-i)*(n-i-1)//2
        assert i >= 0 and j >= 0
        return i,j

    def __eq__(self,other):
        if not hasattr(other, 'dists'):
            return NotImplemented
        
        return self.dists == other.dists

    def __ne__(self,other):
        result = self.__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        return not result

    def __hash__(self):
        return hash(tuple(self.dists))

    def __str__(self):
        return 'DistMatrix(%s)' % str(self.dists)

    def __repr__(self):
        return str(self)

class BasicLinkage:

    log = _getLogger('BasicLinkage')

    def __init__(self, linkage):
        self.lnk = linkage
        self.log.debug('Created with linkage %s', self.lnk)

    def __call__(self,i,j,M):
        self.log.debug('%s-merging (%d,%d)', self.lnk, i, j)
        assert i < j
        linkage = self.lnk
        dists  = [None]*((M.n-1)*(M.n-2)//2)
        lindex = 0
        for s in range(M.n):
            if s == j:
                continue
            for t in range(s+1,M.n):
                if t == j:
                    continue
                elif s == i:
                    dists[lindex] = linkage(i,j,t,M)
                    lindex += 1
                elif t == i:
                    dists[lindex] = linkage(i,j,s,M)
                    lindex += 1
                else:
                    dists[lindex] = M[s,t]
                    lindex += 1

        return DistMatrix(dists=dists)

def getLinkageFactory(L):
    if L == 'single':
        return lambda x : BasicLinkage(SingleLinkage())
    if L == 'complete':
        return lambda x : BasicLinkage(CompleteLinkage())
    if L == 'average':
        return lambda x : BasicLinkage(AverageLinkage(x))

    raise Exception('Unknown linkage model: "%s"' % str(L))
    
def SingleLinkage():
    class SL:
        def __call__(self,i,j,x,M):
            return min(M[i,x],M[j,x])
        def __repr__(self):
            return 'SL'
        def __str__(self):
            return 'SL'
    return SL()

def CompleteLinkage():
    class CL:
        def __call__(self,i,j,x,M):
            return max(M[i,x], M[j,x])
        def __repr__(self):
            return 'CL'
        def __str__(self):
            return 'CL'
    return CL()
    
def AverageLinkage(sizes):
    class AL:
        def __call__(self,i, j, x, M):
            '''
            (i,j) - to be merged
            x     - compute distance to
            M     - old distances
            '''
            a = sizes[i] * M[i,x]
            b = sizes[j] * M[j,x]
            s = sizes[i] + sizes[j]
            return float(a+b)/s

        def __repr__(self):
            return 'AL'
        def __str__(self):
            return 'AL'
    return AL()

class AgglomerativeClustering:
    '''
    Return type for clustering. 
    '''
    def __init__(self,joins=None,dists=None):
        assert (joins is None) == (dists is None)
        if joins:
            assert len(joins) == len(dists)
            self.joins = list(joins)
            self.dists = list(dists)
        else:
            self.joins = []
            self.dists = []

    def isMonotone(self, tolerance=0.0):
        for a,b in zip(self.dists[:-1], self.dists[1:]):
            if b - a < -tolerance:
                return False
        return True

    def __add__(self,other):
        '''
        Concatenation of agglomerative clustering results.
        '''
        joins = list(self.joins)
        joins.extend(other.joins)
        dists = list(self.dists)
        dists.extend(other.dists)
        return AgglomerativeClustering(joins,dists)

    def __eq__(self,other):
        if not isinstance(other, AgglomerativeClustering):
            return False
        return self.joins == other.joins and self.dists == other.dists

    def __ne__(self,other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((hash(tuple(self.joins)),
                     hash(tuple(self.dists))))

    def __getitem__(self, x):
        return AgglomerativeClustering(self.joins[x],self.dists[x])
            

    def __len__(self):
        assert len(self.dists) == len(self.joins)
        return len(self.dists)

    def __str__(self):
        return 'AgglomerativeClustering(joins=%s, dists=%s)' % \
            (str(self.joins), str(self.dists))

    def __repr__(self):
        return str(self)

# Alias
AC = AgglomerativeClustering



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# Copyright 2020 Daniel Bakkelund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

import setuptools
import ophac

with open("README.md", "r") as fh:
    long_description = fh.read()

__version__ = ophac.__version__
    
setuptools.setup(
    name='ophac',
    version=__version__,
    author='Daniel Bakkelund',
    author_email='daniel.bakkelund@ifi.uio.no',
    description='Order Preserving Hierarchical Agglomerative Clustering',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://bitbucket.org/Bakkelund/ophac/src/v04/',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Development Status :: 3 - Alpha',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.0',
    install_requires=[
        'ophac_cpp==0.4.1',
        'numpy',
        'scipy'
    ],
    zip_safe=False
)
